#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from os import PathLike
import asyncio
from pathlib import Path
from fhir.resources.R4B import bundle
from pydicom import dcmread
from pydicom import dataset
from tqdm import tqdm
import logging
from typing import Iterable, Union, AsyncGenerator
import dicom2fhir.helpers as helpers
from dicom2fhir.helpers import get_or
from dicom2fhir.dicom2fhirbundle import Dicom2FHIRBundle
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

StrPath = Union[str, PathLike]

async def async_rglob_files(base_dir: Path, pattern: str = "*") -> AsyncGenerator[Path, None]:
    loop = asyncio.get_running_loop()
    files = await loop.run_in_executor(None, lambda: list(base_dir.rglob(pattern)))
    for file in files:
        if file.is_file():
            yield file

def is_dicom_file(path: Path) -> bool:
      # Ensure the file can be opened in binary mode
    try:
        with path.open('rb') as f:
            f.seek(128)
            return f.read(4) == b'DICM'
    except Exception:
        return False

async def _parse_directory(dcmDir: StrPath, config: dict) -> AsyncGenerator[DicomJsonProxy, None]:
    """
    Parse a directory of DICOM files including subdirectories and return instances as an AsyncGenerator.

    :param dcmDir: Directory containing DICOM files.
    :return: AsyncGenerator[DicomJsonProxy, None]
    """
    base = Path(dcmDir)
    if not base.is_dir():
        raise ValueError(f"Directory '{dcmDir}' not found")

    skip_invalid_files = get_or(config, "directory_parser.skip_invalid_files", True)

    async for f in async_rglob_files(base, pattern="*"):
        if not f.is_file():
            continue
        if skip_invalid_files and not is_dicom_file(f):
            logging.warning(f"Skipping invalid DICOM file: {f}")
            continue

        try:
            ds = dcmread(f, stop_before_pixels=True, force=True)
            yield DicomJsonProxy(ds.to_json_dict())
        except:
            logging.exception(f"An error occurred while processing DICOM file {f}")
            raise

async def _create_bundle(instances: AsyncGenerator[DicomJsonProxy, None], config: dict = {}) -> bundle.Bundle:

    dcm2fhir = Dicom2FHIRBundle(config=config)
    
    async for ds in instances:
        if not isinstance(ds, DicomJsonProxy):
            raise TypeError("Expected a DicomJsonProxy object")
        dcm2fhir.add(ds)

    return dcm2fhir.create_bundle()

async def from_directory(dcms: StrPath, config: dict = {}) -> bundle.Bundle:
    """
    Process DICOM files into an ImagingStudy FHIR resource.
    
    :param dcms: A directory containing DICOM files.
    :return: ImagingStudy resource.
    """

    def _wrap(instances: Iterable[dataset.Dataset]):
        for instance in instances:
            yield DicomJsonProxy(instance.to_json_dict())

    # use default id function for FHIR resource id generation
    if 'id_function' not in config:
        config['id_function'] = helpers.default_id_function()

    # parse directory of DICOM files
    if isinstance(dcms, StrPath):
        if not Path(dcms).is_dir():
            raise ValueError(f"Expected a directory, got: {dcms}")
        dicom_json_proxies = _parse_directory(dcms, config)
        return await _create_bundle(dicom_json_proxies, config)
    # use iterable of DICOM JSON Proxies
    else:
        # guard against non-iterable or non-dict types
        if not isinstance(dcms, Iterable):
            raise TypeError("Expected an iterable of dicts. Got: {}".format(type(dcms)))
        for d in dcms:
            if not isinstance(d, dict):
                raise TypeError("Expected a dict. Got: {}".format(type(d)))

        dicom_json_proxies = [DicomJsonProxy(d) for d in dcms]
        return _create_bundle(dicom_json_proxies, config)

async def from_generator(dcms: AsyncGenerator[dict, None], config: dict = {}) -> bundle.Bundle:
    """
    Process DICOM files or datasets into an ImagingStudy FHIR resource.
    
    :param dcms: AsyncGenerator of DICOM JSON dicts.
    :return: ImagingStudy resource.
    """

    # use default id function for FHIR resource id generation
    if 'id_function' not in config:
        config['id_function'] = helpers.default_id_function()

    # guard against non-iterable or non-dict types
    if not isinstance(dcms, AsyncGenerator):
        raise TypeError("Expected an async generator of dicts. Got: {}".format(type(dcms)))

    async def _to_dicom_json_proxy(dcms: AsyncGenerator[dict, None]) -> AsyncGenerator[DicomJsonProxy, None]:
        async for d in dcms:
            yield DicomJsonProxy(d)

    dicom_json_proxies = _to_dicom_json_proxy(dcms)
    return await _create_bundle(dicom_json_proxies, config)



# -*- coding: utf-8 -*-
import re
import datetime
import uuid
from typing import Union
from pydicom.valuerep import PersonName
from fhir.resources.R4B.patient import Patient
from fhir.resources.R4B.humanname import HumanName
from fhir.resources.R4B.address import Address
from fhir.resources.R4B.contactpoint import ContactPoint
from fhir.resources.R4B.identifier import Identifier
from fhir.resources.R4B.fhirtypes import DateType
from fhir.resources.R4B.extension import Extension
from fhir.resources.R4B.quantity import Quantity
from dicom2fhir.helpers import get_or
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

DATE8_REGEX = re.compile(r'^(19|20)\d{2}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])$')

def dicom_name_to_fhir(name: str) -> HumanName:
    """
    Convert DICOM PersonName to FHIR HumanName.

    DICOM name format: 'Family^Given^Middle^Prefix^Suffix'
    """
    pname = PersonName(name)
    if pname is None:
        return HumanName.model_construct()

    return HumanName.model_construct(
        family      = pname.family_name,
        given       = [n for n in [pname.given_name, pname.middle_name] if n],
        prefix      = [pname.name_prefix] if pname.name_prefix else None,
        suffix      = [pname.name_suffix] if pname.name_suffix else None,
    )


def dicom_birthdate_to_fhir(birthdate: str) -> DateType | None:
     """
     Convert DICOM PatientBirthDate (YYYYMMDD) to FHIR-compliant date.
     Returns a datetime.date or None if invalid.
     """
     if not isinstance(birthdate, str) or not DATE8_REGEX.match(birthdate):
         return None

     try:
         year = int(birthdate[:4])
         month = int(birthdate[4:6])
         day = int(birthdate[6:])
         return datetime.date(year, month, day)
     except ValueError:
         return None

def dicom_gender_to_fhir(sex: str) -> str:
    """
    Map DICOM PatientSex to FHIR gender.
    - 'M' -> 'male'
    - 'F' -> 'female'
    - 'O' or other -> 'other'
    """
    mapping = {
        'M': 'male',
        'F': 'female',
        'O': 'other'
    }
    return mapping.get(str(sex).upper(), 'unknown') if sex else 'unknown'

def dicom_address_to_fhir(address_str: str) -> Address:
    """
    Convert DICOM address string to FHIR Address.
    Format: 'Street^OtherDesignation^City^State^PostalCode^Country'
    """
    if not address_str:
        return Address.model_construct()

    parts = address_str.split("^")
    return Address.model_construct(
        line=[parts[0]] if parts[0] else None,
        city=parts[2] if len(parts) > 2 else None,
        state=parts[3] if len(parts) > 3 else None,
        postalCode=parts[4] if len(parts) > 4 else None,
        country=parts[5] if len(parts) > 5 else None
    )

def build_patient_resource(ds: DicomJsonProxy, config: dict) -> Patient:

    patient = Patient.model_construct()

    # Identifier
    if "PatientID" in ds:

        assigner = None
        issuer_of_patient_id = ds.get("IssuerOfPatientID", None)
        if issuer_of_patient_id is not None and len(str(issuer_of_patient_id)) > 0:
            assigner = {
                "display": str(issuer_of_patient_id)
            }

        patient.id = config['id_function']('Patient', ds)
        patient.identifier = [
            Identifier.model_construct(
                use="usual",
                system="urn:dicom:patient-id",
                value=str(ds.PatientID),
                assigner=assigner
            )
        ]
    else:
        patient.id = str(uuid.uuid4())

    # Name
    if "PatientName" in ds:
        patient.name = [dicom_name_to_fhir(str(ds.PatientName))]

    # BirthDate
    if "PatientBirthDate" in ds:
        patient.birthDate = dicom_birthdate_to_fhir(str(ds.PatientBirthDate))

    # Gender
    if "PatientSex" in ds:
        patient.gender = dicom_gender_to_fhir(str(ds.PatientSex))

    # Address
    if "PatientAddress" in ds:
        patient.address = [dicom_address_to_fhir(str(ds.PatientAddress))]

    # Telecom
    if "PatientTelephoneNumbers" in ds:
        tel = ds.PatientTelephoneNumbers
        if not isinstance(tel, list):
            tel = [tel]
        patient.telecom = [
            ContactPoint.model_construct(system="phone", value=str(phone), use="home") for phone in tel
        ]

    return patient


# -*- coding: utf-8 -*-
import os
import uuid
import hashlib
from typing import Callable
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

def get_or(d: dict, path: str, default=None):
    """
    Get a value from a nested dictionary using a dot-separated path.
    If the path does not exist, return the default value.
    """
    keys = path.split('.')
    val = d
    for key in keys:
        if isinstance(val, dict) and key in val:
            val = val[key]
        else:
            return default
    return val if val is not None else default

def env_or_config(env: str, config_path: str, config: dict):
    """
    Return the value of an environment variable or a configuration key.
    If neither is set raise a ValueError.
    """
    if env in os.environ:
        return os.environ[env]

    val = get_or(config, config_path)
    if val is None:
        raise ValueError(f"Neither environment variable '{env}' nor configuration key '{config_path}' is set.")
    return val

# default id functions
def default_id_function(pepper: str | None = None) -> Callable[[str, DicomJsonProxy], str]:
    """
    Default ID function for FHIR resource id generation.
    Can be customized with a pepper string for additional uniqueness.
    """
    def _id(resource_type: str, ds: DicomJsonProxy, extra: str|None = "") -> str:
        if not isinstance(ds, DicomJsonProxy):
            raise TypeError("Expected a DicomJsonProxy object")

        base_string = extra
        if resource_type == "ImagingStudy" and "StudyInstanceUID" in ds:
            base_string = f"{base_string}{ds.StudyInstanceUID}"
        elif resource_type == "Patient" and "PatientID" in ds:
            base_string = f"{base_string}{ds.PatientID}"
        elif resource_type == "Device" and "DeviceSerialNumber" in ds:
            uid = ds.get("DeviceUID") or ''
            ser = ds.get("DeviceSerialNumber") or ''
            mod = ds.get("ManufacturerModelName") or ''
            base_string = f"{base_string}{uid}{ser}{mod}"
        elif resource_type == "Observation":
            uid = ds.StudyInstanceUID if ds.non_empty("StudyInstanceUID") else ""
            study_date = ds.StudyDate if ds.non_empty("StudyDate") else ""
            study_time = ds.StudyTime if ds.non_empty("StudyTime") else ""
            base_string = f"{base_string}{uid}{study_date}{study_time}"
        else:
            return str(uuid.uuid4())

        return hashlib.sha256(f"{base_string}{pepper or ''}".encode("utf-8")).hexdigest()

    return _id


import uuid
import logging
from collections.abc import Iterable
from dicom2fhir.dicom_json_proxy import DicomJsonProxy
from fhir.resources.R4B.device import Device, DeviceDeviceName
from fhir.resources.R4B.annotation import Annotation
from fhir.resources.R4B.device import DeviceUdiCarrier

logger = logging.getLogger(__name__)

def _map_software_versions(ds: DicomJsonProxy) -> list[dict]:
    """
    Extract SoftwareVersions (DICOM 0018,1020) from the dataset and map to FHIR Device.version.
    
    This attribute is multi-valued (LO), so it returns a list of version strings.
    """
    # pydicom returns either a single value or a MultiValue object for VR LO
    if "SoftwareVersions" not in ds:
        return []
    
    # Normalize to list of strings
    raw = ds.SoftwareVersions
    if isinstance(raw, Iterable) and not isinstance(raw, (str, bytes)):
        return [{'value': str(item).strip()} for item in raw if item is not None]
    else:
        return [{'value': str(raw).strip()}]

def build_device_resource(ds: DicomJsonProxy, config: dict) -> Device:
    """
    Build FHIR Device resource from DICOM metadata (General Equipment Module).
    Extracts manufacturer, model, serial, version, institution, station, calibration, UDI, etc.
    """

    device = Device.model_construct()
    # Resource ID
    device.id = config['id_function']('Device', ds)

    # Identifiers
    identifiers = []

    if ds.non_empty("DeviceSerialNumber"):
        identifiers.append({
            "use": "official",
            "system": "urn:dicom:device-serial-number",
            "value": str(ds.DeviceSerialNumber)
        })

    if ds.non_empty("DeviceUID"):
        identifiers.append({
            "use": "official",
            "system": "urn:dicom:device-uid",
            "value": str(ds.DeviceUID)
        })

    if len(identifiers) > 0:
        device.identifier = identifiers

    # Manufacturer & Model
    if ds.non_empty("Manufacturer"):
        device.manufacturer = str(ds.Manufacturer)
    if ds.non_empty("ManufacturerModelName"):
        device.deviceName = [DeviceDeviceName.model_construct(name=str(ds.ManufacturerModelName), type="model-name")]

    # Software version(s)
    device.version = _map_software_versions(ds)

    # Institutional context
    if ds.non_empty("InstitutionName"):
        device.owner = {"display": str(ds.InstitutionName)}
    if ds.non_empty("InstitutionalDepartmentName"):
        device.location = {"display": str(ds.InstitutionalDepartmentName)}
    if ds.non_empty("StationName"):
        # Could be assigned to device.deviceName as 'station' or use part of location
        device.deviceName = device.deviceName or []
        device.deviceName.append(DeviceDeviceName.model_construct(name=str(ds.StationName), type="station-name"))

    # Physical/device-specific details
    try:
        if ds.non_empty("SpatialResolution"):
            device.property = device.property or []
            device.property.append({
                "type": {"text": "spatial-resolution-mm"},
                "valueQuantity": {"value": float(ds.SpatialResolution), "unit": "mm"}
            })
    except:
        logger.warning(f"Failed to extract SpatialResolution: {ds.SpatialResolution}")

    # Calibration date/time
    dt = ""
    if ds.non_empty("DateOfLastCalibration"):
        dt += str(ds.DateOfLastCalibration) 
    if ds.non_empty("TimeOfLastCalibration"):
        dt += str(ds.TimeOfLastCalibration)
    if dt != "":
        device.note = device.note or []
        device.note.append(Annotation.model_construct(text=f"Last calibration: {dt}"))

    # Pixel padding—maybe not core but included
    if ds.non_empty("PixelPaddingValue"):
        device.note = device.note or []
        device.note.append(Annotation.model_construct(text=f"Pixel padding value: {ds.PixelPaddingValue}"))

    # UDI (Unique Device Identifier)
    if ds.non_empty("UDISequence"):
        udi_items = []
        for item in ds.UDISequence:
            if item.non_empty("UniqueDeviceIdentifier"):
                udi_items.append(DeviceUdiCarrier.model_construct(
                    deviceIdentifier=str(item.UniqueDeviceIdentifier))
                )
        if udi_items:
            device.udiCarrier = udi_items

    # Modality as device type
    if ds.non_empty("Modality"):
        device.type = {"text": str(ds.Modality)}

    return device


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pandas as pd

BODYSITE_SNOMED_MAPPING_URL = "https://dicom.nema.org/medical/dicom/current/output/chtml/part16/chapter_L.html"

def download_body_part_mapping(url=BODYSITE_SNOMED_MAPPING_URL):
    print(f"Downloading: {url}")
    dfs = pd.read_html(url, converters={"Code Value": str})
    df = dfs[2][["Code Value", "Code Meaning", "Body Part Examined"]]
    df = df.dropna(subset=["Body Part Examined"])
    return df

def save_json(df, name):
    out_dir = os.path.join("resources", "terminologies")
    os.makedirs(out_dir, exist_ok=True)
    path = os.path.join(out_dir, f"{name}.json")
    df.to_json(path, orient="records", force_ascii=False)
    print(f"Saved {len(df)} entries to {path}")

def main():
    df_bp = download_body_part_mapping()
    save_json(df_bp, "bodysite_snomed")
    print("All terminologies built.")

if __name__ == "__main__":
    main()


from datetime import datetime
from dateutil import tz as dateutil_tz

from fhir.resources.R4B import imagingstudy
from fhir.resources.R4B import identifier
from fhir.resources.R4B import codeableconcept
from fhir.resources.R4B import coding
from fhir.resources.R4B import patient
from fhir.resources.R4B import humanname
from fhir.resources.R4B import fhirtypes
from fhir.resources.R4B import reference
from fhir.resources.R4B import extension
import pandas as pd
import json
from pathlib import Path
import os
import logging
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

TERMINOLOGY_CODING_SYS = "http://terminology.hl7.org/CodeSystem/v2-0203"
TERMINOLOGY_CODING_SYS_CODE_ACCESSION = "ACSN"
TERMINOLOGY_CODING_SYS_CODE_MRN = "MR"

ACQUISITION_MODALITY_SYS = "http://dicom.nema.org/resources/ontology/DCM"
SCANNING_SEQUENCE_SYS = "https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.3.html"
SCANNING_VARIANT_SYS = "https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.3.html"

SOP_CLASS_SYS = "urn:ietf:rfc:3986"

# load rather expesive resource into global var to make it reusable
BODYSITE_SNOMED_MAPPING_PATH = Path(__file__).parent / "resources" / "terminologies" / "bodysite_snomed.json"
BODYSITE_SNOMED_MAPPING = pd.DataFrame(json.loads(BODYSITE_SNOMED_MAPPING_PATH.read_text(encoding="utf-8")))

def _get_snomed(dicom_bodypart: str, sctmapping: pd.DataFrame) -> dict[str, str] | None:
    _rec = sctmapping.loc[sctmapping['Body Part Examined'] == dicom_bodypart]
    if _rec.empty:
        return None
    return {
        'code': _rec["Code Value"].iloc[0],
        'display': _rec["Code Meaning"].iloc[0],
    }

def gen_accession_identifier(id):
    idf = identifier.Identifier.model_construct()
    idf.use = "usual"
    idf.type = codeableconcept.CodeableConcept.model_construct()
    idf.type.coding = []
    acsn = coding.Coding.model_construct()
    acsn.system = TERMINOLOGY_CODING_SYS
    acsn.code = TERMINOLOGY_CODING_SYS_CODE_ACCESSION

    idf.type.coding.append(acsn)
    idf.value = id
    return idf


def gen_studyinstanceuid_identifier(id):
    idf = identifier.Identifier.model_construct()
    idf.system = "urn:dicom:uid"
    idf.value = "urn:oid:" + id
    return idf


def get_patient_resource_ids(PatientID, IssuerOfPatientID):
    idf = identifier.Identifier.model_construct()
    idf.use = "usual"
    idf.value = str(PatientID)

    idf.type = codeableconcept.CodeableConcept.model_construct()
    idf.type.coding = []
    id_coding = coding.Coding.model_construct()
    id_coding.system = TERMINOLOGY_CODING_SYS
    id_coding.code = TERMINOLOGY_CODING_SYS_CODE_MRN
    idf.type.coding.append(id_coding)

    if IssuerOfPatientID is not None:
        idf.assigner = reference.Reference.model_construct()
        idf.assigner.display = str(IssuerOfPatientID)

    return idf


def calc_gender(gender: str | None):
    if gender is None:
        return "unknown"
    if not gender:
        return "unknown"
    if gender.upper().lower() == "f":
        return "female"
    if gender.upper().lower() == "m":
        return "male"
    if gender.upper().lower() == "o":
        return "other"

    return "unknown"


def calc_dob(dicom_dob: str):
    if dicom_dob == '':
        return None

    try:
        dob = datetime.strptime(dicom_dob, '%Y%m%d')
        fhir_dob = fhirtypes.Date(
            dob.year,
            dob.month,
            dob.day
        )
    except Exception:
        return None
    return fhir_dob


def inline_patient_resource(referenceId, PatientID, IssuerOfPatientID, patientName, gender, dob):
    p = patient.Patient.model_construct()
    p.id = referenceId
    p.name = []
    # p.use = "official"
    p.identifier = [get_patient_resource_ids(PatientID, IssuerOfPatientID)]
    hn = humanname.HumanName.model_construct()
    hn.family = str(patientName.family_name)
    if patientName.given_name != '':
        hn.given = [str(patientName.given_name)]
    p.name.append(hn)
    p.gender = calc_gender(str(gender))
    p.birthDate = calc_dob(str(dob))
    p.active = True
    return p


def gen_procedurecode_array(procedures):
    if procedures is None:
        return None
    fhir_proc = []
    for p in procedures:
        concept = codeableconcept.CodeableConcept.model_construct()
        c = coding.Coding.model_construct()
        c.system = p["system"]
        c.code = p["code"]
        c.display = p["display"]
        concept.coding = []
        concept.coding.append(c)
        concept.text = p["display"]
        fhir_proc.append(concept)
    if len(fhir_proc) > 0:
        return fhir_proc
    return None


def gen_started_datetime(dt, tm, tz):
    """
    Generate a timezone-aware datetime object from DICOM date and time strings.

    Args:
        dt (str): DICOM date in the format 'YYYYMMDD'.
        tm (str): DICOM time in the format 'HHMMSS' or shorter.
        tz (str): Timezone as a string (e.g., 'Europe/Berlin' or '+01:00').

    Returns:
        datetime: A timezone-aware datetime object or None.
    """
    if dt is None:
        return None

    dt_pattern = '%Y%m%d'
    if tm is not None and len(tm) >= 6:
        studytm = datetime.strptime(tm[0:6], '%H%M%S')
        dt_string = f"{dt} {studytm.hour:02d}:{studytm.minute:02d}:{studytm.second:02d}"
        dt_pattern += " %H:%M:%S"
    else:
        dt_string = dt

    try:
        dt_date = datetime.strptime(dt_string, dt_pattern)
    except ValueError:
        return None

    # Apply timezone
    try:
        if tz:
            tzinfo = dateutil_tz.gettz(tz)
            if tzinfo is not None:
                dt_date = dt_date.replace(tzinfo=tzinfo)
    except Exception:
        pass

    return dt_date

def gen_reason(reason, reasonStr):
    if reason is None and reasonStr is None:
        return None
    reasonList = []
    if reason is None or len(reason) <= 0:
        # Only assign if non-empty and not just whitespace
        if reasonStr and reasonStr.strip():
            rc = codeableconcept.CodeableConcept.model_construct()
            rc.text = reasonStr
            reasonList.append(rc)
        return reasonList

    for r in reason:
        rc = codeableconcept.CodeableConcept.model_construct()
        rc.coding = []
        c = coding.Coding.model_construct()
        c.system = r["system"]
        c.code = r["code"]
        c.display = r["display"]
        rc.coding.append(c)
        reasonList.append(rc)
    return reasonList


def gen_coding(code: str, system: str|None = None, display: str|None = None):
    if isinstance(code, list):
        raise Exception(
        "More than one code for type Coding detected")
    c = coding.Coding.model_construct()
    c.code = code
    c.system = system
    c.display = display
    if system is None and display is None:
        c.userSelected = True

    return c

def gen_codeable_concept(value_list: list, system):
    c = codeableconcept.CodeableConcept.model_construct()
    c.coding = []
    for _l in value_list:
        m = gen_coding(_l, system)
        c.coding.append(m)
    return c

def gen_bodysite_coding(bd):

    bd_snomed = _get_snomed(bd, sctmapping=BODYSITE_SNOMED_MAPPING)
    
    if bd_snomed is None:
        return gen_coding(code=str(bd))
    
    return gen_coding(
        code=str(bd_snomed['code']),
        system="http://snomed.info/sct",
        display=bd_snomed['display']
    )

def dcm_coded_concept(code_sequence: list[DicomJsonProxy]):
    """    
    Convert a DICOM Code Sequence to a list of FHIR CodeableConcept objects.
    Args:
        code_sequence (list[DicomJsonProxy]): A list of DicomJsonProxy objects representing the DICOM Code Sequence.
    Returns:
        list[dict]: A list of dictionaries representing the FHIR CodeableConcept objects.
    Raises:
        TypeError: If the input is not a list of DicomJsonProxy objects.
    """

    if not isinstance(code_sequence, list):
        raise TypeError("Expected a list of DicomJsonProxy objects")

    concepts = []
    for seq in code_sequence:

        if not isinstance(seq, DicomJsonProxy):
            raise TypeError("Expected a DicomJsonProxy object in the list")

        concept = {}
        if seq.non_empty("CodeValue"):
            concept["code"] = str(seq.CodeValue)
        if seq.non_empty("CodingSchemeDesignator"):
            concept["system"] = str(seq.CodingSchemeDesignator)
        if seq.non_empty("CodeMeaning"):
            concept["display"] = str(seq.CodeMeaning)
        concepts.append(concept)
    return concepts


# -*- coding: utf-8 -*-
import logging
from typing import List
from fhir.resources.R4B.observation import Observation
from fhir.resources.R4B.patient import Patient
from fhir.resources.R4B.imagingstudy import ImagingStudy
from fhir.resources.R4B.quantity import Quantity
from fhir.resources.R4B.codeableconcept import CodeableConcept
from fhir.resources.R4B.coding import Coding
from fhir.resources.R4B.reference import Reference
from dicom2fhir.dicom2fhirutils import gen_started_datetime
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

logger = logging.getLogger(__name__)

def build_observation_resources(ds: DicomJsonProxy, patient: Patient, study: ImagingStudy, config: dict) -> List[Observation]:
    observations = []
    
    def create_obs(code: str, display: str, value: float, unit: str, system: str, code_unit: str) -> Observation:
        return Observation.model_construct(
            id=config['id_function']('Observation', ds, extra=code),
            status="final",
            category=[CodeableConcept.model_construct(
                coding=[Coding.model_construct(system="http://terminology.hl7.org/CodeSystem/observation-category", code="vital-signs")]
            )],
            code=CodeableConcept.model_construct(
                coding=[Coding.model_construct(system="http://loinc.org", code=code, display=display)],
                text=display
            ),
            subject=Reference.model_construct(reference=f"Patient/{patient.id}") if patient else None,
            partOf=[Reference.model_construct(reference=f"ImagingStudy/{study.id}")] if study else [],
            effectiveDateTime=gen_started_datetime(ds.StudyDate, ds.StudyTime, config["dicom_timezone"]),
            valueQuantity=Quantity.model_construct(value=value, unit=unit, system=system, code=code_unit)
        )

    if "PatientWeight" in ds and ds.PatientWeight is not None:
        try:
            weight = float(ds.PatientWeight)
            observations.append(create_obs("29463-7", "Body Weight", weight, "kg", "http://unitsofmeasure.org", "kg"))
        except:
            logger.warning(f"Failed to extract PatientWeight: {ds.PatientWeight}")

    if "PatientSize" in ds and ds.PatientSize is not None:
        try:
            height = float(ds.PatientSize)
            observations.append(create_obs("8302-2", "Body Height", height, "m", "http://unitsofmeasure.org", "m"))
        except:
            logger.warning(f"Failed to extract PatientSize: {ds.PatientSize}")

    return observations


import json
from pydicom.datadict import tag_for_keyword
from pydicom.tag import Tag

class DicomJsonProxy:
    def __init__(self, dicom_json: dict):
        self._raw = dicom_json

    def __repr__(self):
        return json.dumps(self._raw, indent=2)

    def __str__(self):
        return repr(self)

    def __getattr__(self, tag_name):
        tag = tag_for_keyword(tag_name)
        if tag is None:
            raise AttributeError(f"Unknown DICOM keyword: {tag_name}")
        key = f"{tag >> 16:04X}{tag & 0xFFFF:04X}"
        try:
            return self._extract_value(self._raw[key])
        except KeyError:
            raise AttributeError(f"No such DICOM attribute: {tag_name}")

    def __getitem__(self, tag):
        if tag in self._raw:
            return self._extract_value(self._raw[tag])
        raise KeyError(f"Tag {tag} not found in DICOM JSON.")

    def __contains__(self, tag_or_keyword):
        # Handle keyword like "PatientID"
        if isinstance(tag_or_keyword, str):
            tag = tag_for_keyword(tag_or_keyword)
            if tag is None:
                return False
            tag_key = f"{tag >> 16:04X}{tag & 0xFFFF:04X}"
        # Handle int or Tag
        elif isinstance(tag_or_keyword, (int, Tag)):
            tag = Tag(tag_or_keyword)
            tag_key = f"{tag.group:04X}{tag.element:04X}"
        else:
            return False
        return tag_key in self._raw

    def _extract_value(self, elem):

        def _wrap_as_proxy(value):
            if isinstance(value, dict):
                return DicomJsonProxy(value)
            elif isinstance(value, list):
                return [DicomJsonProxy(item) if isinstance(item, dict) else item for item in value]
            return value

        if isinstance(elem, dict) and "Value" in elem and "vr" in elem:
            if elem["vr"] == "SQ":
                return _wrap_as_proxy(elem["Value"]) or []
            return _wrap_as_proxy(elem["Value"][0]) if isinstance(elem["Value"], list) else _wrap_as_proxy(elem["Value"])
        return _wrap_as_proxy(elem)

    def get(self, name, default=None):
        try:
            return getattr(self, name)
        except AttributeError:
            return default
        
    def non_empty(self, name):
        """
        Check if the DICOM attribute is present and not empty.
        """
        value = self.get(name)

        # Handle cases where value is a dict or list
        if isinstance(value, dict):
            if "Value" in value:
                value = value["Value"]
            else:
                return False
        if isinstance(value, list):
            return len(value) > 0

        return value is not None and str(value).strip() != ''


import os
import yaml, json
import logging
import unittest
from unittest import skipUnless
from pathlib import Path
from .. import dicom2fhir
from fhir.resources.R4B import bundle
from fhir.resources.R4B import imagingstudy
from .. import helpers
from dicom2fhir.dicom_json_proxy import DicomJsonProxy
from typing import AsyncGenerator
from pydicom import dcmread

dicom2fhir_config = {
    "dicom_timezone": "Europe/Berlin",
    "generator": {
        "imaging_study": {
            "add_instance": True,
        }
    }
}

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

def _extract_imaging_study_from_bundle(b: bundle.Bundle) -> imagingstudy.ImagingStudy:
    """
    Extract the ImagingStudy resource from a FHIR Bundle.
    """
    if not isinstance(b, bundle.Bundle):
        raise TypeError("Expected a Bundle resource")

    for entry in (b.entry or []):
        if entry.resource and isinstance(entry.resource, imagingstudy.ImagingStudy):
            return entry.resource

    raise ValueError("No ImagingStudy resource found in the bundle")

class testDicom2FHIR(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        # Set up any necessary configuration or environment variables
        with open(Path(__file__).parent / "config.yaml") as f:
            self.config = yaml.safe_load(f)

        if not self.config:
            raise ValueError("Configuration could not be loaded")

    async def test_dicom_json_proxy(self):

        dcm_file = os.path.join(os.getcwd(), "dicom2fhir", "tests", "resources", "dcm-instance", "dcm_1.dcm")
        dicom_json = dcmread(dcm_file, stop_before_pixels=True, force=True).to_json_dict()
        dicom_json_proxy = DicomJsonProxy(dicom_json)

        pat_id = dicom_json_proxy.PatientID
        self.assertIsNotNone(pat_id, "PatientID should not be None")
        self.assertIsInstance(pat_id, str, "PatientID should be a string")
        self.assertTrue(pat_id == "DAC007_CRLAT", "PatientID should be DAC007_CRLAT")

        kvp = dicom_json_proxy.KVP

        logger.info(f"KVP: {kvp}, type: {type(kvp)}")

  


    async def test_instance_dicom2fhir(self):
        dcmDir = os.path.join(os.getcwd(), "dicom2fhir", "tests", "resources", "dcm-instance")
        study: imagingstudy.ImagingStudy
        bundle = await dicom2fhir.from_directory(dcmDir, config=self.config)
        study = _extract_imaging_study_from_bundle(bundle)

        self.assertIsNotNone(study, "No ImagingStudy was generated")
        self.assertEqual(study.numberOfSeries, 1, "Number of Series in the study mismatch")
        self.assertEqual(study.numberOfInstances, 1, "Number of Instances in the study mismatch")
        self.assertIsNotNone(study.series, "Series was not built for the study")
        self.assertIsNotNone(study.modality, "Modality is missing")
        self.assertEqual(len(study.modality), 1, "Series must list only one modality")
        self.assertEqual(study.modality[0].code, "CR", "Incorrect modality detected")
        self.assertEqual(len(study.series), 1, "Number objects in Series Array: mismatch")
        self.assertEqual(len(study.series[0].instance), 1, "Number objects in Instance Array: mismatch")

        series: imagingstudy.ImagingStudySeries
        series = study.series[0]
        self.assertIsNotNone(series, "Missing Series")
        self.assertIsNotNone(series.bodySite, "Body site is missing")
        self.assertEqual(series.bodySite.code, '43799004', "Expected SNOMED code for CHEST")
        self.assertIsNotNone(series.bodySite.display, "BodySite display is missing")
        self.assertEqual(series.bodySite.display, 'Chest', "Chest is expected as body site")

        instance: imagingstudy.ImagingStudySeriesInstance
        instance = series.instance[0]
        self.assertIsNotNone(instance, "Missing Instance")

    async def test_multi_instance_dicom(self):
        dcmDir = os.path.join(os.getcwd(), "dicom2fhir", "tests", "resources", "dcm-multi-instance")
        bundle = await dicom2fhir.from_directory(dcmDir, config=self.config)

        #print(bundle.model_dump_json(indent=2))

        study = _extract_imaging_study_from_bundle(bundle)

        self.assertIsNotNone(study, "No ImagingStudy was generated")
        self.assertEqual(study.numberOfSeries, 1)
        self.assertEqual(study.numberOfInstances, 5)
        self.assertIsNotNone(study.series, "Series was not built for the study")
        self.assertEqual(len(study.modality), 1, "Only single modality expected for this study")
        self.assertEqual(study.modality[0].code, "CR")
        self.assertEqual(len(study.series), 1, "Incorrect number of series detected")
        self.assertEqual(len(study.series[0].instance), 5, "Incorrect number of instances detected")

    async def test_multi_series_dicom(self):
        dcmDir = os.path.join(os.getcwd(), "dicom2fhir", "tests", "resources", "dcm-multi-series")
        bundle = await dicom2fhir.from_directory(dcmDir, config=self.config)

        #print(bundle.model_dump_json(indent=2))

        study = _extract_imaging_study_from_bundle(bundle)

        self.assertIsNotNone(study, "No ImagingStudy was generated")
        self.assertEqual(study.numberOfSeries, 4, "Number of Series in the study mismatch")
        self.assertEqual(study.numberOfInstances, 4, "Number of Instances in the study mismatch")
        self.assertIsNotNone(study.series, "Series was not built for the study")
        self.assertEqual(len(study.modality), 1, "Only single modality expected for this study")
        self.assertEqual(study.modality[0].code, "CR", "Incorrect Modality detected")
        self.assertEqual(len(study.series), 4, "Number of series in the study: mismatch")

    @skipUnless(os.getenv("RUN_FMX_TESTS") == "1", "Skipping FMX tests by config")
    async def test_fmx(self):
        import asyncpg

        def load_fmx_conn_params(config: dict) -> dict:
            """
            Load connection params from YAML and inject password from ENV.
            """
            if os.getenv("FMX_PASSWORD") is None:
                raise ValueError("FMX_PASSWORD environment variable is not set.")

            return {
                "database": helpers.env_or_config("FMX_DATABASE", "fmx.database", config),
                "user":  helpers.env_or_config("FMX_USER", "fmx.user", config),
                "host": helpers.env_or_config("FMX_HOST", "fmx.host", config),
                "port": int(helpers.env_or_config("FMX_PORT", "fmx.port", config)),
                "password": os.getenv("FMX_PASSWORD")
            }

        conn_params = load_fmx_conn_params(self.config)

        async with asyncpg.create_pool(**conn_params) as pool:
            async with pool.acquire() as conn:
                async with conn.transaction():
                    for _ in range(10):
                        stmt = await conn.prepare("""
                            SELECT di.sop_instance_uid, di.tags
                            FROM dicom.dicom_instances di
                            WHERE di.study_instance_uid = (
                                SELECT study_instance_uid
                                FROM dicom.dicom_studies
                                ORDER BY random()
                                LIMIT 1
                            )
                        """)

                        async def _records_generator():
                            async for record in stmt.cursor():
                                yield json.loads(record["tags"])

                        bundle = await dicom2fhir.from_generator(_records_generator(), config=self.config)

                        study = _extract_imaging_study_from_bundle(bundle)

                        self.assertIsNotNone(study, "No ImagingStudy was generated")
                        self.assertIsNotNone(study.series, "Series was not built for the study")
                        self.assertIsNotNone(study.series[0].instance[0].sopClass, "SOP Class is missing")
                        self.assertEqual(study.series[0].instance[0].sopClass.system, "urn:ietf:rfc:3986", "SOP Class system is incorrect")


import uuid
from fhir.resources.R4B import bundle
from fhir.resources.R4B import imagingstudy
from fhir.resources.R4B import patient
from fhir.resources.R4B import device
from fhir.resources.R4B.reference import Reference
#from pydicom import dataset
import logging
from dicom2fhir.dicom2fhirutils import gen_coding, SOP_CLASS_SYS, ACQUISITION_MODALITY_SYS, gen_bodysite_coding, gen_accession_identifier, gen_studyinstanceuid_identifier, dcm_coded_concept, gen_procedurecode_array, gen_started_datetime, gen_reason
from dicom2fhir.dicom2patient import build_patient_resource
from dicom2fhir.dicom2observation import build_observation_resources
from dicom2fhir.dicom2device import build_device_resource
from dicom2fhir.helpers import get_or
from dicom2fhir.dicom_json_proxy import DicomJsonProxy

class Dicom2FHIRBundle():

    def __init__(self, config: dict = {}):
        """
        Initialize the Dicom2FHIRBundle with an optional configuration.
        """
        self.study: imagingstudy.ImagingStudy | None = None
        self.series = {}
        self.instances = {}
        # Device
        self.device: device.Device | None = None
        # Patient
        self.pat: patient.Patient | None = None
        self.obs = []
        self.config = config
        
    def add(self, ds: DicomJsonProxy):
        """
        Add a DICOM dataset to the ImagingStudy.
        """
        if not isinstance(ds, DicomJsonProxy):
            raise TypeError("Expected a DicomJsonProxy object")

        # is first instance?
        if self.study is None:
            self.device = build_device_resource(ds, self.config)
            self.pat = build_patient_resource(ds, self.config)
            self._create_imaging_study(ds)
            if get_or(self.config, "generator.observation.add_vital_signs", True):
                self.obs = build_observation_resources(ds, self.pat, self.study, self.config) or []

        self._add_imaging_study_series(ds)
        self._add_instance(ds)

    def _create_imaging_study(self, ds):

        study_data = {}
        study_data["resource_type"] = "ImagingStudy"
        study_data["id"] = self.config['id_function']("ImagingStudy", ds)
        study_data["status"] = "available"
        
        if ds.non_empty("StudyDescription"):
            study_data["description"] = str(ds.StudyDescription)
        
        study_data["identifier"] = []
        if ds.non_empty("AccessionNumber"):
            study_data["identifier"].append(gen_accession_identifier(str(ds.AccessionNumber)))
        if ds.non_empty("StudyInstanceUID"):
            study_data["identifier"].append(gen_studyinstanceuid_identifier(str(ds.StudyInstanceUID)))

        # Set the patient reference
        study_data["subject"] = Reference.model_construct(reference=f"Patient/{self.pat.id}") if self.pat else None

        # procedure codes
        if ds.non_empty("ProcedureCodeSequence"):
            procedures = dcm_coded_concept(ds.ProcedureCodeSequence)
            study_data["procedureCode"] = gen_procedurecode_array(procedures)        

        if ds.non_empty("StudyDate") and ds.non_empty("StudyTime"):
            study_data["started"] = gen_started_datetime(str(ds.StudyDate), str(ds.StudyTime), self.config["dicom_timezone"])

        # reason codes
        if ds.non_empty("ReferencedRequestSequence"):
            for seq in ds.ReferencedRequestSequence:

                reason = None
                reasonStr = None 

                if seq.non_empty("ReasonForRequestedProcedureCodeSequence"):
                    reason = dcm_coded_concept(ds.ReasonForRequestedProcedureCodeSequence)
                if seq.non_empty("ReasonForTheRequestedProcedure"):
                    reasonStr = str(ds.ReasonForTheRequestedProcedure)
                if reason is not None and reasonStr is not None:
                    study_data["reasonCode"] = gen_reason(reason, reasonStr)

        study_data["numberOfSeries"] = 0
        study_data["numberOfInstances"] = 0

        # instantiate study here, when all required fields are available
        self.study = imagingstudy.ImagingStudy(**study_data)
        
    def _add_imaging_study_series(self, ds: DicomJsonProxy):

        series_instance_uid = str(ds.SeriesInstanceUID)

        if series_instance_uid in self.series:
            return  # Series already exists

        # inti data container
        self.series[series_instance_uid] = {}
        self.series[series_instance_uid]["uid"] = series_instance_uid
        self.series[series_instance_uid]["numberOfInstances"] = 0

        if ds.non_empty("SeriesDescription"):
            self.series[series_instance_uid]["description"] = str(ds.SeriesDescription)

        if ds.non_empty("SeriesNumber"):
            self.series[series_instance_uid]["number"] = str(ds.SeriesNumber)

        if ds.non_empty("Modality"):
            self.series[series_instance_uid]["modality"] = gen_coding(
                code=str(ds.Modality),
                system=ACQUISITION_MODALITY_SYS
            )
        
        if ds.non_empty("SeriesDate") and ds.non_empty("SeriesTime"):
            self.series[series_instance_uid]["started"] = gen_started_datetime(
                str(ds.SeriesDate), str(ds.SeriesTime), self.config["dicom_timezone"]
            )

        if ds.non_empty("BodyPartExamined"):
            self.series[series_instance_uid]["bodySite"] = gen_bodysite_coding(str(ds.BodyPartExamined))

        if ds.non_empty("Laterality"):
            self.series[series_instance_uid]["laterality"] = gen_coding(str(ds.Laterality))
    
    def _add_instance(self, ds: DicomJsonProxy):

        series_instance_uid = str(ds.SeriesInstanceUID)
        sop_instance_uid = str(ds.SOPInstanceUID)

        if series_instance_uid not in self.instances:
            self.instances[series_instance_uid] = {}

        if sop_instance_uid in self.instances[series_instance_uid]:
            print("Error: SOP Instance UID already exists in the series")
            print(self.instances[series_instance_uid][sop_instance_uid].as_json())
            return

        self.instances[series_instance_uid][sop_instance_uid] = {}
       
        self.instances[series_instance_uid][sop_instance_uid]["uid"] = sop_instance_uid
        self.instances[series_instance_uid][sop_instance_uid]["sopClass"] = gen_coding(
            code="urn:oid:" + ds.SOPClassUID,
            system=SOP_CLASS_SYS
        )

        if ds.non_empty("InstanceNumber"):
            self.instances[series_instance_uid][sop_instance_uid]["number"] = str(ds.InstanceNumber)

        try:
            if str(ds.Modality) == "SR":
                seq = ds.ConceptNameCodeSequence
                if isinstance(seq, list) and len(seq) > 0:
                    self.instances[series_instance_uid][sop_instance_uid]["title"] = str(seq[0].CodeMeaning)
            else:
                self.instances[series_instance_uid][sop_instance_uid]["title"] = '\\'.join(str(ds.ImageType))
        except Exception:
            pass  # print("Unable to set instance title")

    def _build_imaging_study(self) -> imagingstudy.ImagingStudy:
        """
        Build the ImagingStudy resource from the collected data.
        """
        if self.study is None:
            raise ValueError("No ImagingStudy data has been added")

        # Add series to the study
        self.study.series = []
        for series_uid, series_data in self.series.items():

            # only add sop instance data if configured
            if get_or(self.config, "generator.imaging_study.add_instances", False):
                series_data["instance"] = []
                for instance_uid, instance_data in self.instances.get(series_uid, {}).items():
                    series_data["instance"].append(imagingstudy.ImagingStudySeriesInstance(**instance_data))
                series_data["numberOfInstances"] = len(series_data["instance"])
            else:
                series_data["numberOfInstances"] = len(self.instances[series_uid])

            # Create ImagingStudySeries object
            series = imagingstudy.ImagingStudySeries(**series_data)
            self.study.series.append(series)

        # Set the number of series and instances
        self.study.numberOfSeries = len(self.study.series)
        self.study.numberOfInstances = sum(s.numberOfInstances for s in self.study.series)

        # Set the modalities
        modality_set = {
            s.modality.code: s.modality
            for s in self.study.series or []
            if s.modality is not None
        }
        self.study.modality = list(modality_set.values())

        return self.study
    
    def create_bundle(self) -> bundle.Bundle:
        """
        Create the final transaction bundle.
        """

        def _to_entry(resource):
            return {
                #'fullUrl': f"urn:uuid:{resource.id}",
                'resource': resource,
                'request': {
                    'method': 'PUT',
                    'url': f"{resource.__resource_type__}/{resource.id}"
                }
            }

        if not self.study:
            raise ValueError("No ImagingStudy data has been added")

        # Build the ImagingStudy resource
        _study = self._build_imaging_study()
    
        # wrap entries in a transaction Bundle and return
        return bundle.Bundle.model_validate({
            'resourceType': 'Bundle',
            'type': "transaction",
            'id': str(uuid.uuid4()),
            'entry': [
                _to_entry(_study),
                _to_entry(self.pat),
                _to_entry(self.device)
            ] + [_to_entry(o) for o in self.obs]
        })